@@ -31,13 +31,17 @@ Transition matrices
     transition_cycle
     transition_local
 """
+from __future__ import annotations
 
 import numpy as np
 from scipy.spatial.distance import cdist
 from numba import jit
-from .util import pad_center, fill_off_diagonal, tiny, expand_to
+from .util import pad_center, fill_off_diagonal, is_positive_int, tiny, expand_to
 from .util.exceptions import ParameterError
 from .filters import get_window
+from typing import Any, Iterable, List, Optional, Tuple, Union, overload
+from typing_extensions import Literal
+from ._typing import _WindowSpec
 
 __all__ = [
     "dtw",
@@ -53,21 +57,163 @@ __all__ = [
 ]
 
 
+@overload
 def dtw(
-    X=None,
-    Y=None,
+    X: np.ndarray,
+    Y: np.ndarray,
     *,
-    C=None,
-    metric="euclidean",
-    step_sizes_sigma=None,
-    weights_add=None,
-    weights_mul=None,
-    subseq=False,
-    backtrack=True,
-    global_constraints=False,
-    band_rad=0.25,
-    return_steps=False,
-):
+    metric: str = ...,
+    step_sizes_sigma: Optional[np.ndarray] = ...,
+    weights_add: Optional[np.ndarray] = ...,
+    weights_mul: Optional[np.ndarray] = ...,
+    subseq: bool = ...,
+    backtrack: Literal[False],
+    global_constraints: bool = ...,
+    band_rad: float = ...,
+    return_steps: Literal[False] = ...,
+) -> np.ndarray:
+    ...
+
+
+@overload
+def dtw(
+    *,
+    C: np.ndarray,
+    metric: str = ...,
+    step_sizes_sigma: Optional[np.ndarray] = ...,
+    weights_add: Optional[np.ndarray] = ...,
+    weights_mul: Optional[np.ndarray] = ...,
+    subseq: bool = ...,
+    backtrack: Literal[False],
+    global_constraints: bool = ...,
+    band_rad: float = ...,
+    return_steps: Literal[False] = ...,
+) -> np.ndarray:
+    ...
+
+
+@overload
+def dtw(
+    X: np.ndarray,
+    Y: np.ndarray,
+    *,
+    metric: str = ...,
+    step_sizes_sigma: Optional[np.ndarray] = ...,
+    weights_add: Optional[np.ndarray] = ...,
+    weights_mul: Optional[np.ndarray] = ...,
+    subseq: bool = ...,
+    backtrack: Literal[False],
+    global_constraints: bool = ...,
+    band_rad: float = ...,
+    return_steps: Literal[True],
+) -> Tuple[np.ndarray, np.ndarray]:
+    ...
+
+
+@overload
+def dtw(
+    *,
+    C: np.ndarray,
+    metric: str = ...,
+    step_sizes_sigma: Optional[np.ndarray] = ...,
+    weights_add: Optional[np.ndarray] = ...,
+    weights_mul: Optional[np.ndarray] = ...,
+    subseq: bool = ...,
+    backtrack: Literal[False],
+    global_constraints: bool = ...,
+    band_rad: float = ...,
+    return_steps: Literal[True],
+) -> Tuple[np.ndarray, np.ndarray]:
+    ...
+
+
+@overload
+def dtw(
+    X: np.ndarray,
+    Y: np.ndarray,
+    *,
+    metric: str = ...,
+    step_sizes_sigma: Optional[np.ndarray] = ...,
+    weights_add: Optional[np.ndarray] = ...,
+    weights_mul: Optional[np.ndarray] = ...,
+    subseq: bool = ...,
+    backtrack: Literal[True] = ...,
+    global_constraints: bool = ...,
+    band_rad: float = ...,
+    return_steps: Literal[False] = ...,
+) -> Tuple[np.ndarray, np.ndarray]:
+    ...
+
+
+@overload
+def dtw(
+    *,
+    C: np.ndarray,
+    metric: str = ...,
+    step_sizes_sigma: Optional[np.ndarray] = ...,
+    weights_add: Optional[np.ndarray] = ...,
+    weights_mul: Optional[np.ndarray] = ...,
+    subseq: bool = ...,
+    backtrack: Literal[True] = ...,
+    global_constraints: bool = ...,
+    band_rad: float = ...,
+    return_steps: Literal[False] = ...,
+) -> Tuple[np.ndarray, np.ndarray]:
+    ...
+
+
+@overload
+def dtw(
+    X: np.ndarray,
+    Y: np.ndarray,
+    *,
+    metric: str = ...,
+    step_sizes_sigma: Optional[np.ndarray] = ...,
+    weights_add: Optional[np.ndarray] = ...,
+    weights_mul: Optional[np.ndarray] = ...,
+    subseq: bool = ...,
+    backtrack: Literal[True] = ...,
+    global_constraints: bool = ...,
+    band_rad: float = ...,
+    return_steps: Literal[True],
+) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
+    ...
+
+
+@overload
+def dtw(
+    *,
+    C: np.ndarray,
+    metric: str = ...,
+    step_sizes_sigma: Optional[np.ndarray] = ...,
+    weights_add: Optional[np.ndarray] = ...,
+    weights_mul: Optional[np.ndarray] = ...,
+    subseq: bool = ...,
+    backtrack: Literal[True] = ...,
+    global_constraints: bool = ...,
+    band_rad: float = ...,
+    return_steps: Literal[True],
+) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
+    ...
+
+
+def dtw(
+    X: Optional[np.ndarray] = None,
+    Y: Optional[np.ndarray] = None,
+    *,
+    C: Optional[np.ndarray] = None,
+    metric: str = "euclidean",
+    step_sizes_sigma: Optional[np.ndarray] = None,
+    weights_add: Optional[np.ndarray] = None,
+    weights_mul: Optional[np.ndarray] = None,
+    subseq: bool = False,
+    backtrack: bool = True,
+    global_constraints: bool = False,
+    band_rad: float = 0.25,
+    return_steps: bool = False,
+) -> Union[
+    np.ndarray, Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]
+]:
     """Dynamic time warping (DTW).
 
     This function performs a DTW and path backtracking on two sequences.
@@ -97,7 +243,7 @@ def dtw(
         in `scipy.spatial.distance.cdist()`
 
     step_sizes_sigma : np.ndarray [shape=[n, 2]]
-        Specifies allowed step sizes as used by the dtw.
+        Specifies allowed step sizes as used by the DTW.
 
     weights_add : np.ndarray [shape=[n, ]]
         Additive weights to penalize certain step sizes.
@@ -126,12 +272,14 @@ def dtw(
     -------
     D : np.ndarray [shape=(N, M)]
         accumulated cost matrix.
-        D[N, M] is the total alignment cost.
-        When doing subsequence DTW, D[N,:] indicates a matching function.
-    wp : np.ndarray [shape=(N, 2)]
+        The value at the final index position ``D[-1, -1]`` is the total alignment cost.
+    wp : np.ndarray [shape=(L, 2)]
         Warping path with index pairs.
         Each row of the array contains an index pair (n, m).
         Only returned when ``backtrack`` is True.
+        Note that the length ``L`` of the warping path need not match the
+        lengths of the input data, depending on the ``step_sizes_sigma`` values
+        and ``subseq``.
     steps : np.ndarray [shape=(N, M)]
         Step matrix, containing the indices of the used steps from the cost
         accumulation step.
@@ -200,6 +348,11 @@ def dtw(
         weights_add = np.concatenate((default_weights_add, weights_add))
         weights_mul = np.concatenate((default_weights_mul, weights_mul))
 
+    # These asserts are bad, but mypy cannot trace the code paths properly
+    assert step_sizes_sigma is not None
+    assert weights_add is not None
+    assert weights_mul is not None
+
     if np.any(step_sizes_sigma < 0):
         raise ParameterError("step_sizes_sigma cannot contain negative values")
 
@@ -221,14 +374,17 @@ def dtw(
     C_local = False
     if C is None:
         C_local = True
+        # mypy can't figure out that this case does not happen
+        assert X is not None and Y is not None
         # take care of dimensions
         X = np.atleast_2d(X)
         Y = np.atleast_2d(Y)
 
         # Perform some shape-squashing here
         # Put the time axes around front
-        X = np.swapaxes(X, -1, 0)
-        Y = np.swapaxes(Y, -1, 0)
+        # Suppress types because mypy doesn't know these are ndarrays
+        X = np.swapaxes(X, -1, 0)  # type: ignore
+        Y = np.swapaxes(Y, -1, 0)  # type: ignore
 
         # Flatten the remaining dimensions
         # Use F-ordering to preserve columns
@@ -295,6 +451,8 @@ def dtw(
     steps[:, 0] = 2
 
     # calculate accumulated cost matrix
+    D: np.ndarray
+    steps: np.ndarray
     D, steps = __dtw_calc_accu_cost(
         C, D, steps, step_sizes_sigma, weights_mul, weights_add, max_0, max_1
     )
@@ -303,7 +461,9 @@ def dtw(
     D = D[max_0:, max_1:]
     steps = steps[max_0:, max_1:]
 
+    return_values: List[np.ndarray]
     if backtrack:
+        wp: np.ndarray
         if subseq:
             if np.all(np.isinf(D[-1])):
                 raise ParameterError(
@@ -311,7 +471,7 @@ def dtw(
                     "be constructed with the given step sizes."
                 )
             start = np.argmin(D[-1, :])
-            wp = __dtw_backtracking(steps, step_sizes_sigma, subseq, start)
+            _wp = __dtw_backtracking(steps, step_sizes_sigma, subseq, start)
         else:
             # perform warping path backtracking
             if np.isinf(D[-1, -1]):
@@ -320,14 +480,14 @@ def dtw(
                     "be constructed with the given step sizes."
                 )
 
-            wp = __dtw_backtracking(steps, step_sizes_sigma, subseq)
-            if wp[-1] != (0, 0):
+            _wp = __dtw_backtracking(steps, step_sizes_sigma, subseq)
+            if _wp[-1] != (0, 0):
                 raise ParameterError(
                     "Unable to compute a full DTW warping path. "
                     "You may want to try again with subseq=True."
                 )
 
-        wp = np.asarray(wp, dtype=int)
+        wp = np.asarray(_wp, dtype=int)
 
         # since we transposed in the beginning, we have to adjust the index pairs back
         if subseq and (
@@ -344,15 +504,24 @@ def dtw(
         return_values.append(steps)
 
     if len(return_values) > 1:
-        return tuple(return_values)
+        # Suppressing type check here because mypy can't
+        # infer the exact length of the tuple
+        return tuple(return_values)  # type: ignore
     else:
         return return_values[0]
 
 
-@jit(nopython=True, cache=True)
+@jit(nopython=True, cache=True)  # type: ignore
 def __dtw_calc_accu_cost(
-    C, D, steps, step_sizes_sigma, weights_mul, weights_add, max_0, max_1
-):  # pragma: no cover
+    C: np.ndarray,
+    D: np.ndarray,
+    steps: np.ndarray,
+    step_sizes_sigma: np.ndarray,
+    weights_mul: np.ndarray,
+    weights_add: np.ndarray,
+    max_0: int,
+    max_1: int,
+) -> Tuple[np.ndarray, np.ndarray]:  # pragma: no cover
     """Calculate the accumulated cost matrix D.
 
     Use dynamic programming to calculate the accumulated costs.
@@ -415,8 +584,13 @@ def __dtw_calc_accu_cost(
     return D, steps
 
 
-@jit(nopython=True, cache=True)
-def __dtw_backtracking(steps, step_sizes_sigma, subseq, start=None):  # pragma: no cover
+@jit(nopython=True, cache=True)  # type: ignore
+def __dtw_backtracking(
+    steps: np.ndarray,
+    step_sizes_sigma: np.ndarray,
+    subseq: bool,
+    start: Optional[int] = None,
+) -> List[Tuple[int, int]]:  # pragma: no cover
     """Backtrack optimal warping path.
 
     Uses the saved step sizes from the cost accumulation
@@ -479,7 +653,13 @@ def __dtw_backtracking(steps, step_sizes_sigma, subseq, start=None):  # pragma:
     return wp
 
 
-def dtw_backtracking(steps, *, step_sizes_sigma=None, subseq=False, start=None):
+def dtw_backtracking(
+    steps: np.ndarray,
+    *,
+    step_sizes_sigma: Optional[np.ndarray] = None,
+    subseq: bool = False,
+    start: Optional[Union[int, np.integer[Any]]] = None,
+) -> np.ndarray:
     """Backtrack a warping path.
 
     Uses the saved step sizes from the cost accumulation
@@ -510,8 +690,7 @@ def dtw_backtracking(steps, *, step_sizes_sigma=None, subseq=False, start=None):
     """
     if subseq is False and start is not None:
         raise ParameterError(
-            "start is only allowed to be set if subseq is True "
-            "(start={}, subseq={})".format(start, subseq)
+            f"start is only allowed to be set if subseq is True (start={start}, subseq={subseq})"
         )
 
     # Default Parameters
@@ -528,7 +707,50 @@ def dtw_backtracking(steps, *, step_sizes_sigma=None, subseq=False, start=None):
     return np.asarray(wp, dtype=int)
 
 
-def rqa(sim, *, gap_onset=1, gap_extend=1, knight_moves=True, backtrack=True):
+@overload
+def rqa(
+    sim: np.ndarray,
+    *,
+    gap_onset: float = ...,
+    gap_extend: float = ...,
+    knight_moves: bool = ...,
+    backtrack: Literal[False],
+) -> np.ndarray:
+    ...
+
+
+@overload
+def rqa(
+    sim: np.ndarray,
+    *,
+    gap_onset: float = ...,
+    gap_extend: float = ...,
+    knight_moves: bool = ...,
+    backtrack: Literal[True] = ...,
+) -> Tuple[np.ndarray, np.ndarray]:
+    ...
+
+
+@overload
+def rqa(
+    sim: np.ndarray,
+    *,
+    gap_onset: float = ...,
+    gap_extend: float = ...,
+    knight_moves: bool = ...,
+    backtrack: bool = ...,
+) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
+    ...
+
+
+def rqa(
+    sim: np.ndarray,
+    *,
+    gap_onset: float = 1,
+    gap_extend: float = 1,
+    knight_moves: bool = True,
+    backtrack: bool = True,
+) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
     """Recurrence quantification analysis (RQA)
 
     This function implements different forms of RQA as described by
@@ -656,12 +878,13 @@ def rqa(sim, *, gap_onset=1, gap_extend=1, knight_moves=True, backtrack=True):
     >>> ax[1].legend()
     >>> ax[1].label_outer()
     """
-
     if gap_onset < 0:
         raise ParameterError("gap_onset={} must be strictly positive")
     if gap_extend < 0:
         raise ParameterError("gap_extend={} must be strictly positive")
 
+    score: np.ndarray
+    pointers: np.ndarray
     score, pointers = __rqa_dp(sim, gap_onset, gap_extend, knight_moves)
     if backtrack:
         path = __rqa_backtrack(score, pointers)
@@ -670,10 +893,11 @@ def rqa(sim, *, gap_onset=1, gap_extend=1, knight_moves=True, backtrack=True):
     return score
 
 
-@jit(nopython=True, cache=True)
-def __rqa_dp(sim, gap_onset, gap_extend, knight):  # pragma: no cover
+@jit(nopython=True, cache=True)  # type: ignore
+def __rqa_dp(
+    sim: np.ndarray, gap_onset: float, gap_extend: float, knight: bool
+) -> Tuple[np.ndarray, np.ndarray]:  # pragma: no cover
     """RQA dynamic programming implementation"""
-
     # The output array
     score = np.zeros(sim.shape, dtype=sim.dtype)
 
@@ -688,7 +912,7 @@ def __rqa_dp(sim, gap_onset, gap_extend, knight):  # pragma: no cover
     #
     # If knight moves are disabled, then only the first entry is used.
     #
-    # Using dummy vectors here makes the code a bit cleaner down below.
+    # Using placeholder vectors here makes the code a bit cleaner down below.
     sim_values = np.zeros(3)
     score_values = np.zeros(3)
     vec = np.zeros(3)
@@ -829,7 +1053,6 @@ def __rqa_backtrack(score, pointers):
     Given the score matrix and backtracking index array,
     reconstruct the optimal path.
     """
-
     # backtracking rubric:
     #   0 ==> diagonal move
     #   1 ==> knight move up
@@ -845,7 +1068,7 @@ def __rqa_backtrack(score, pointers):
     idx = list(np.unravel_index(np.argmax(score), score.shape))
 
     # Construct the path
-    path = []
+    path: List = []
     while True:
         bt_index = pointers[tuple(idx)]
 
@@ -875,8 +1098,10 @@ def __rqa_backtrack(score, pointers):
     return np.asarray(path, dtype=np.uint)
 
 
-@jit(nopython=True, cache=True)
-def _viterbi(log_prob, log_trans, log_p_init):  # pragma: no cover
+@jit(nopython=True, cache=True)  # type: ignore
+def _viterbi(
+    log_prob: np.ndarray, log_trans: np.ndarray, log_p_init: np.ndarray
+) -> Tuple[np.ndarray, np.ndarray]:  # pragma: no cover
     """Core Viterbi algorithm.
 
     This is intended for internal use only.
@@ -937,7 +1162,35 @@ def _viterbi(log_prob, log_trans, log_p_init):  # pragma: no cover
     return state, logp
 
 
-def viterbi(prob, transition, *, p_init=None, return_logp=False):
+@overload
+def viterbi(
+    prob: np.ndarray,
+    transition: np.ndarray,
+    *,
+    p_init: Optional[np.ndarray] = ...,
+    return_logp: Literal[True],
+) -> Tuple[np.ndarray, np.ndarray]:
+    ...
+
+
+@overload
+def viterbi(
+    prob: np.ndarray,
+    transition: np.ndarray,
+    *,
+    p_init: Optional[np.ndarray] = ...,
+    return_logp: Literal[False] = ...,
+) -> np.ndarray:
+    ...
+
+
+def viterbi(
+    prob: np.ndarray,
+    transition: np.ndarray,
+    *,
+    p_init: Optional[np.ndarray] = None,
+    return_logp: bool = False,
+) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
     """Viterbi decoding from observation likelihoods.
 
     Given a sequence of observation likelihoods ``prob[s, t]``,
@@ -1011,13 +1264,12 @@ def viterbi(prob, transition, *, p_init=None, return_logp=False):
     >>> print(logp, path)
     -4.19173690823075 [0 0 1]
     """
-
     n_states, n_steps = prob.shape[-2:]
 
     if transition.shape != (n_states, n_states):
         raise ParameterError(
-            "transition.shape={}, must be "
-            "(n_states, n_states)={}".format(transition.shape, (n_states, n_states))
+            f"transition.shape={transition.shape}, must be "
+            f"(n_states, n_states)={n_states, n_states}"
         )
 
     if np.any(transition < 0) or not np.allclose(transition.sum(axis=1), 1):
@@ -1040,9 +1292,7 @@ def viterbi(prob, transition, *, p_init=None, return_logp=False):
         or not np.allclose(p_init.sum(), 1)
         or p_init.shape != (n_states,)
     ):
-        raise ParameterError(
-            "Invalid initial state distribution: " "p_init={}".format(p_init)
-        )
+        raise ParameterError(f"Invalid initial state distribution: p_init={p_init}")
 
     log_trans = np.log(transition + epsilon)
     log_prob = np.log(prob + epsilon)
@@ -1054,6 +1304,9 @@ def viterbi(prob, transition, *, p_init=None, return_logp=False):
         # Transpose outputs for return
         return _state.T, logp
 
+    states: np.ndarray
+    logp: np.ndarray
+
     if log_prob.ndim == 2:
         states, logp = _helper(log_prob)
     else:
@@ -1073,9 +1326,50 @@ def viterbi(prob, transition, *, p_init=None, return_logp=False):
     return states
 
 
+@overload
 def viterbi_discriminative(
-    prob, transition, *, p_state=None, p_init=None, return_logp=False
-):
+    prob: np.ndarray,
+    transition: np.ndarray,
+    *,
+    p_state: Optional[np.ndarray] = ...,
+    p_init: Optional[np.ndarray] = ...,
+    return_logp: Literal[False] = ...,
+) -> np.ndarray:
+    ...
+
+
+@overload
+def viterbi_discriminative(
+    prob: np.ndarray,
+    transition: np.ndarray,
+    *,
+    p_state: Optional[np.ndarray] = ...,
+    p_init: Optional[np.ndarray] = ...,
+    return_logp: Literal[True],
+) -> Tuple[np.ndarray, np.ndarray]:
+    ...
+
+
+@overload
+def viterbi_discriminative(
+    prob: np.ndarray,
+    transition: np.ndarray,
+    *,
+    p_state: Optional[np.ndarray] = ...,
+    p_init: Optional[np.ndarray] = ...,
+    return_logp: bool,
+) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
+    ...
+
+
+def viterbi_discriminative(
+    prob: np.ndarray,
+    transition: np.ndarray,
+    *,
+    p_state: Optional[np.ndarray] = None,
+    p_init: Optional[np.ndarray] = None,
+    return_logp: bool = False,
+) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
     """Viterbi decoding from discriminative state predictions.
 
     Given a sequence of conditional state predictions ``prob[s, t]``,
@@ -1091,6 +1385,11 @@ def viterbi_discriminative(
     ``P[State(t) = s | Obs(t)] / P[State(t) = s]``, where the denominator
     is the marginal probability of state ``s`` occurring as given by ``p_state``.
 
+    Note that because the denominator ``P[State(t) = s]`` is not explicitly
+    calculated, the resulting probabilities (or log-probabilities) are not
+    normalized.  If using the `return_logp=True` option (see below),
+    be aware that the "probabilities" may not sum to (and may exceed) 1.
+
     Parameters
     ----------
     prob : np.ndarray [shape=(..., n_states, n_steps), non-negative]
@@ -1118,8 +1417,8 @@ def viterbi_discriminative(
         If ``prob`` contains multiple input channels,
         then each channel is decoded independently.
     logp : scalar [float] or np.ndarray
-        If ``return_logp=True``, the log probability of ``states`` given
-        the observations.
+        If ``return_logp=True``, the (unnormalized) log probability
+        of ``states`` given the observations.
 
     See Also
     --------
@@ -1186,13 +1485,12 @@ def viterbi_discriminative(
     ...        yticklabels=[labels[i] for i in np.unique(chords_vit)])
     >>> ax.legend()
     """
-
     n_states, n_steps = prob.shape[-2:]
 
     if transition.shape != (n_states, n_states):
         raise ParameterError(
-            "transition.shape={}, must be "
-            "(n_states, n_states)={}".format(transition.shape, (n_states, n_states))
+            f"transition.shape={transition.shape}, must be "
+            f"(n_states, n_states)={n_states, n_states}"
         )
 
     if np.any(transition < 0) or not np.allclose(transition.sum(axis=1), 1):
@@ -1217,12 +1515,10 @@ def viterbi_discriminative(
     elif p_state.shape != (n_states,):
         raise ParameterError(
             "Marginal distribution p_state must have shape (n_states,). "
-            "Got p_state.shape={}".format(p_state.shape)
+            f"Got p_state.shape={p_state.shape}"
         )
     elif np.any(p_state < 0) or not np.allclose(p_state.sum(axis=-1), 1):
-        raise ParameterError(
-            "Invalid marginal state distribution: " "p_state={}".format(p_state)
-        )
+        raise ParameterError(f"Invalid marginal state distribution: p_state={p_state}")
 
     if p_init is None:
         p_init = np.empty(n_states)
@@ -1232,9 +1528,7 @@ def viterbi_discriminative(
         or not np.allclose(p_init.sum(), 1)
         or p_init.shape != (n_states,)
     ):
-        raise ParameterError(
-            "Invalid initial state distribution: " "p_init={}".format(p_init)
-        )
+        raise ParameterError(f"Invalid initial state distribution: p_init={p_init}")
 
     # By Bayes' rule, P[X | Y] * P[Y] = P[Y | X] * P[X]
     # P[X] is constant for the sake of maximum likelihood inference
@@ -1258,6 +1552,8 @@ def viterbi_discriminative(
         # Transpose outputs for return
         return _state.T, logp
 
+    states: np.ndarray
+    logp: np.ndarray
     if log_prob.ndim == 2:
         states, logp = _helper(log_prob)
     else:
@@ -1277,7 +1573,50 @@ def viterbi_discriminative(
     return states
 
 
-def viterbi_binary(prob, transition, *, p_state=None, p_init=None, return_logp=False):
+@overload
+def viterbi_binary(
+    prob: np.ndarray,
+    transition: np.ndarray,
+    *,
+    p_state: Optional[np.ndarray] = ...,
+    p_init: Optional[np.ndarray] = ...,
+    return_logp: Literal[False] = ...,
+) -> np.ndarray:
+    ...
+
+
+@overload
+def viterbi_binary(
+    prob: np.ndarray,
+    transition: np.ndarray,
+    *,
+    p_state: Optional[np.ndarray] = ...,
+    p_init: Optional[np.ndarray] = ...,
+    return_logp: Literal[True],
+) -> Tuple[np.ndarray, np.ndarray]:
+    ...
+
+
+@overload
+def viterbi_binary(
+    prob: np.ndarray,
+    transition: np.ndarray,
+    *,
+    p_state: Optional[np.ndarray] = ...,
+    p_init: Optional[np.ndarray] = ...,
+    return_logp: bool = ...,
+) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
+    ...
+
+
+def viterbi_binary(
+    prob: np.ndarray,
+    transition: np.ndarray,
+    *,
+    p_state: Optional[np.ndarray] = None,
+    p_init: Optional[np.ndarray] = None,
+    return_logp: bool = False,
+) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
     """Viterbi decoding from binary (multi-label), discriminative state predictions.
 
     Given a sequence of conditional state predictions ``prob[s, t]``,
@@ -1295,6 +1634,10 @@ def viterbi_binary(prob, transition, *, p_state=None, p_init=None, return_logp=F
     The output is a binary matrix ``states[s, t]`` indicating whether each
     state ``s`` is active at time ``t``.
 
+    Like `viterbi_discriminative`, the probabilities of the optimal state sequences
+    are not normalized here.  If using the `return_logp=True` option (see below),
+    be aware that the "probabilities" may not sum to (and may exceed) 1.
+
     Parameters
     ----------
     prob : np.ndarray [shape=(..., n_steps,) or (..., n_states, n_steps)], non-negative
@@ -1325,7 +1668,7 @@ def viterbi_binary(prob, transition, *, p_state=None, p_init=None, return_logp=F
         If not provided, it is assumed to be uniform.
 
     return_logp : bool
-        If ``True``, return the log-likelihood of the state sequence.
+        If ``True``, return the (unnormalized) log-likelihood of the state sequences.
 
     Returns
     -------
@@ -1333,8 +1676,8 @@ def viterbi_binary(prob, transition, *, p_state=None, p_init=None, return_logp=F
     states : np.ndarray [shape=(..., n_states, n_steps)]
         The most likely state sequence.
     logp : np.ndarray [shape=(..., n_states,)]
-        If ``return_logp=True``, the log probability of each state activation
-        sequence ``states``
+        If ``return_logp=True``, the (unnormalized) log probability of each
+        state activation sequence ``states``
 
     See Also
     --------
@@ -1361,7 +1704,6 @@ def viterbi_binary(prob, transition, *, p_state=None, p_init=None, return_logp=F
     >>> librosa.sequence.viterbi_binary(prob, trans, p_state=0.5, p_init=0.5)
     array([[0, 0, 0, 0, 1, 1, 1, 0, 0, 0]])
     """
-
     prob = np.atleast_2d(prob)
 
     n_states, n_steps = prob.shape[-2:]
@@ -1370,8 +1712,8 @@ def viterbi_binary(prob, transition, *, p_state=None, p_init=None, return_logp=F
         transition = np.tile(transition, (n_states, 1, 1))
     elif transition.shape != (n_states, 2, 2):
         raise ParameterError(
-            "transition.shape={}, must be (2, 2) or "
-            "(n_states, 2, 2)={}".format(transition.shape, (n_states))
+            f"transition.shape={transition.shape}, must be (2, 2) or "
+            f"(n_states, 2, 2)={n_states}"
         )
 
     if np.any(transition < 0) or not np.allclose(transition.sum(axis=-1), 1):
@@ -1389,10 +1731,10 @@ def viterbi_binary(prob, transition, *, p_state=None, p_init=None, return_logp=F
     else:
         p_state = np.atleast_1d(p_state)
 
+    assert p_state is not None
+
     if p_state.shape != (n_states,) or np.any(p_state < 0) or np.any(p_state > 1):
-        raise ParameterError(
-            "Invalid marginal state distributions: p_state={}".format(p_state)
-        )
+        raise ParameterError(f"Invalid marginal state distributions: p_state={p_state}")
 
     if p_init is None:
         p_init = np.empty(n_states)
@@ -1400,10 +1742,10 @@ def viterbi_binary(prob, transition, *, p_state=None, p_init=None, return_logp=F
     else:
         p_init = np.atleast_1d(p_init)
 
+    assert p_init is not None
+
     if p_init.shape != (n_states,) or np.any(p_init < 0) or np.any(p_init > 1):
-        raise ParameterError(
-            "Invalid initial state distributions: p_init={}".format(p_init)
-        )
+        raise ParameterError(f"Invalid initial state distributions: p_init={p_init}")
 
     shape_prefix = list(prob.shape[:-2])
     states = np.empty(shape_prefix + [n_states, n_steps], dtype=np.uint16)
@@ -1437,7 +1779,7 @@ def viterbi_binary(prob, transition, *, p_state=None, p_init=None, return_logp=F
     return states
 
 
-def transition_uniform(n_states):
+def transition_uniform(n_states: int) -> np.ndarray:
     """Construct a uniform transition matrix over ``n_states``.
 
     Parameters
@@ -1457,16 +1799,15 @@ def transition_uniform(n_states):
            [0.333, 0.333, 0.333],
            [0.333, 0.333, 0.333]])
     """
-
-    if not isinstance(n_states, (int, np.integer)) or n_states <= 0:
-        raise ParameterError("n_states={} must be a positive integer")
+    if not is_positive_int(n_states):
+        raise ParameterError(f"n_states={n_states} must be a positive integer")
 
     transition = np.empty((n_states, n_states), dtype=np.float64)
     transition.fill(1.0 / n_states)
     return transition
 
 
-def transition_loop(n_states, prob):
+def transition_loop(n_states: int, prob: Union[float, Iterable[float]]) -> np.ndarray:
     """Construct a self-loop transition matrix over ``n_states``.
 
     The transition matrix will have the following properties:
@@ -1505,9 +1846,8 @@ def transition_loop(n_states, prob):
            [0.25 , 0.5  , 0.25 ],
            [0.375, 0.375, 0.25 ]])
     """
-
-    if not isinstance(n_states, (int, np.integer)) or n_states <= 1:
-        raise ParameterError("n_states={} must be a positive integer > 1")
+    if not (is_positive_int(n_states) and (n_states > 1)):
+        raise ParameterError(f"n_states={n_states} must be a positive integer > 1")
 
     transition = np.empty((n_states, n_states), dtype=np.float64)
 
@@ -1519,13 +1859,11 @@ def transition_loop(n_states, prob):
 
     if prob.shape != (n_states,):
         raise ParameterError(
-            "prob={} must have length equal to n_states={}".format(prob, n_states)
+            f"prob={prob} must have length equal to n_states={n_states}"
         )
 
     if np.any(prob < 0) or np.any(prob > 1):
-        raise ParameterError(
-            "prob={} must have values in the range [0, 1]".format(prob)
-        )
+        raise ParameterError(f"prob={prob} must have values in the range [0, 1]")
 
     for i, prob_i in enumerate(prob):
         transition[i] = (1.0 - prob_i) / (n_states - 1)
@@ -1534,7 +1872,7 @@ def transition_loop(n_states, prob):
     return transition
 
 
-def transition_cycle(n_states, prob):
+def transition_cycle(n_states: int, prob: Union[float, Iterable[float]]) -> np.ndarray:
     """Construct a cyclic transition matrix over ``n_states``.
 
     The transition matrix will have the following properties:
@@ -1572,9 +1910,8 @@ def transition_cycle(n_states, prob):
            [0. , 0. , 0.9, 0.1],
            [0.1, 0. , 0. , 0.9]])
     """
-
-    if not isinstance(n_states, (int, np.integer)) or n_states <= 1:
-        raise ParameterError("n_states={} must be a positive integer > 1")
+    if not (is_positive_int(n_states) and n_states > 1):
+        raise ParameterError(f"n_states={n_states} must be a positive integer > 1")
 
     transition = np.zeros((n_states, n_states), dtype=np.float64)
 
@@ -1586,13 +1923,11 @@ def transition_cycle(n_states, prob):
 
     if prob.shape != (n_states,):
         raise ParameterError(
-            "prob={} must have length equal to n_states={}".format(prob, n_states)
+            f"prob={prob} must have length equal to n_states={n_states}"
         )
 
     if np.any(prob < 0) or np.any(prob > 1):
-        raise ParameterError(
-            "prob={} must have values in the range [0, 1]".format(prob)
-        )
+        raise ParameterError(f"prob={prob} must have values in the range [0, 1]")
 
     for i, prob_i in enumerate(prob):
         transition[i, np.mod(i + 1, n_states)] = 1.0 - prob_i
@@ -1601,7 +1936,13 @@ def transition_cycle(n_states, prob):
     return transition
 
 
-def transition_local(n_states, width, *, window="triangle", wrap=False):
+def transition_local(
+    n_states: int,
+    width: Union[int, Iterable[int]],
+    *,
+    window: _WindowSpec = "triangle",
+    wrap: bool = False,
+) -> np.ndarray:
     """Construct a localized transition matrix.
 
     The transition matrix will have the following properties:
@@ -1673,9 +2014,8 @@ def transition_local(n_states, width, *, window="triangle", wrap=False):
            [0.   , 0.   , 0.333, 0.333, 0.333],
            [0.   , 0.   , 0.   , 0.   , 1.   ]])
     """
-
-    if not isinstance(n_states, (int, np.integer)) or n_states <= 1:
-        raise ParameterError("n_states={} must be a positive integer > 1")
+    if not (is_positive_int(n_states) and n_states > 1):
+        raise ParameterError(f"n_states={n_states} must be a positive integer > 1")
 
     width = np.asarray(width, dtype=int)
     if width.ndim == 0:
@@ -1683,11 +2023,11 @@ def transition_local(n_states, width, *, window="triangle", wrap=False):
 
     if width.shape != (n_states,):
         raise ParameterError(
-            "width={} must have length equal to n_states={}".format(width, n_states)
+            f"width={width} must have length equal to n_states={n_states}"
         )
 
     if np.any(width < 1):
-        raise ParameterError("width={} must be at least 1")
+        raise ParameterError(f"width={width} must be at least 1")
 
     transition = np.zeros((n_states, n_states), dtype=np.float64)
 
